/*
 * Depends on libpam-dev
 * unix_chkpwd.c
 *
 * This program is designed to run setuid(root) or with sufficient
 * privilege to access any PAM services. It is designed
 * to provide a mechanism for the current user (defined by this
 * process' uid) to verify their own password.
 *
 * The password is read from the standard input. The exit status of
 * this program indicates whether the user is authenticated or not.
 */
/*
 * Copyright (c) 2000, 2006 Fabian Knittel
 *
 * The file is based on unix_chkpwd, which was
 * Copyright (c) 1996 Andrew G. Morgan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <string.h>
#include <syslog.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#include <signal.h>
#include <security/pam_appl.h>

#define MAXPASS    	200	/* the maximum length of a password */

/**
 * Structure for application specific data passed through PAM
 * to our conv call-back routine chkpwd_pam_conv.
 */
typedef struct {
    const char *login;    		/* plaintext authenticator */
    const char *password;    	/* plaintext password */
    pam_handle_t *pamh;    		/* pointer to PAM handle */
} pam_appdata;

/**
 * Call-back function used by the PAM library to communicate with us. Each
 * received message expects a response, pointed to by resp.
 */
static int chkpwd_pam_conv(
  int num_msg,                      /* I: number of messages */
  const struct pam_message **msg,   /* I: pointer to array of messages */
  struct pam_response **resp,    	/* O: pointer to pointer of response */
  void *appdata_ptr                 /* I: pointer to app specific data */
  )
{
    pam_appdata *my_appdata;    	/* application specific data */
    struct pam_response *my_resp;   /* response created by this func */
    int i;    		            	/* loop counter */
    const char *login_prompt;    	/* string prompting for user-name */
    int rc;    			            /* return code holder */

    my_appdata = appdata_ptr;

    my_resp = malloc(sizeof(struct pam_response) * num_msg);
    if (my_resp == NULL)
    return PAM_CONV_ERR;

    for (i = 0; i < num_msg; i++) {
	    switch (msg[i]->msg_style) {
		    /*
		     * We assume PAM_PROMPT_ECHO_OFF to be a request for password.
		     * This assumption might be unsafe.
		     *
		     * For PAM_PROMPT_ECHO_ON we first check whether the provided
		     * request string matches PAM_USER_PROMPT and, only if they do
		     * match, assume it to be a request for the login.
		     */
		    case PAM_PROMPT_ECHO_OFF:	/* password */
		        my_resp[i].resp = strdup(my_appdata->password);
		        if (my_resp[i].resp == NULL) {
			    	syslog(LOG_DEBUG, "DEBUG: chkpwd_pam_conv: "
                           "strdup failed");
			    	goto ret_error;
		        }
		        my_resp[i].resp_retcode = PAM_SUCCESS;
		        break;
		
		    case PAM_PROMPT_ECHO_ON:	/* username? */
		        /* Recheck setting each time, as it might have been changed
		           in the mean-while. */
		        rc = pam_get_item(my_appdata->pamh, PAM_USER_PROMPT,
		            		      (void *) &login_prompt);
		        if (rc != PAM_SUCCESS) {
			    	syslog(LOG_DEBUG, "DEBUG: chkpwd_pam_conv: unable to "
                           "read login prompt string: %s",
			    	       pam_strerror(my_appdata->pamh, rc));
			    	goto ret_error;
		        }
		
		        if (strcmp(msg[i]->msg, login_prompt) == 0) {
			    	my_resp[i].resp = strdup(my_appdata->login);
			    	my_resp[i].resp_retcode = PAM_SUCCESS;
		        } else {			/* ignore */
			    	syslog(LOG_DEBUG, "DEBUG: chkpwd_pam_conv: unknown "
                           "prompt string: %s", msg[i]->msg);
			    	my_resp[i].resp = NULL;
			    	my_resp[i].resp_retcode = PAM_SUCCESS;
		        }
		        break;
		
		    case PAM_ERROR_MSG:		/* ignore */
		    case PAM_TEXT_INFO:		/* ignore */
		        my_resp[i].resp = NULL;
		        my_resp[i].resp_retcode = PAM_SUCCESS;
		        break;
		
		    default:			/* error */
		        goto ret_error;
	    }
    }
    *resp = my_resp;
    return PAM_SUCCESS;

ret_error:
    /*
     * Free response structure. Don't free my_resp[i], as that
     * isn't initialised yet.
     */
    {
    int y;

    for (y = 0; y < i; y++)
        if (my_resp[y].resp != NULL)
            free(my_resp[y].resp);
        free(my_resp);
    }
    return PAM_CONV_ERR;
}

static int auth_pam(const char *login, const char *password)
{
    pam_appdata my_appdata;    	/* application specific data */
    struct pam_conv my_conv;    	/* pam conversion data */
    pam_handle_t *pamh;    		/* pointer to PAM handle */
    int rc;    			/* return code holder */

    my_appdata.login = login;
    my_appdata.password = password;
    my_appdata.pamh = NULL;

    my_conv.conv = chkpwd_pam_conv;
    my_conv.appdata_ptr = &my_appdata;

    rc = pam_start("chkpwd", login, &my_conv, &pamh);
    if (rc != PAM_SUCCESS) {
        syslog(LOG_DEBUG, "DEBUG: auth_pam: pam_start failed: %s",
               pam_strerror(pamh, rc));
        return PAM_AUTH_ERR;
    }

    my_appdata.pamh = pamh;

    rc = pam_authenticate(pamh, PAM_SILENT);
    if (rc != PAM_SUCCESS) {
        syslog(LOG_DEBUG, "DEBUG: auth_pam: pam_authenticate failed: %s",
               pam_strerror(pamh, rc));
        pam_end(pamh, rc);
        return PAM_AUTH_ERR;
    }

    syslog(LOG_DEBUG, "DEBUG: auth_pam: pam_authenticate successful for %s",
           login);
    pam_end(pamh, PAM_SUCCESS);
    return PAM_SUCCESS;
}

static void my_sighandler(int sig)
{
    if (sig > 0) {
    	syslog(LOG_NOTICE, "caught signal %d.", sig);
    	exit(sig);
    }
}

static void setup_signals(void)
{
    struct sigaction action;	/* posix signal structure */
    /*
     * Setup signal handlers
     */

    (void) memset((void *) &action, 0, sizeof(action));
    action.sa_handler = my_sighandler;
    action.sa_flags = SA_RESETHAND;

    (void) sigaction(SIGILL, &action, NULL);
    (void) sigaction(SIGTRAP, &action, NULL);
    (void) sigaction(SIGBUS, &action, NULL);
    (void) sigaction(SIGSEGV, &action, NULL);

    action.sa_handler = SIG_IGN;
    action.sa_flags = 0;
    (void) sigaction(SIGTERM, &action, NULL);
    (void) sigaction(SIGHUP, &action, NULL);
    (void) sigaction(SIGINT, &action, NULL);
    (void) sigaction(SIGQUIT, &action, NULL);
}

int main(int argc, char *argv[])
{
    char pass[MAXPASS + 1];
    char *option;
    int npass;
    int retval = PAM_AUTH_ERR;
    struct passwd *pw;
    static char user[32];

    /*
     * Catch or ignore as many signal as possible.
     */
    setup_signals();

    /*
     * We establish that this program is running with non-tty stdin.
     * This is to discourage casual use. It does *NOT* prevent an
     * intruder from repeatadly running this program to determine the
     * password of the current user (brute force attack, but one for
     * which the attacker must already have gained access to the user's
     * account).
     */
    if (isatty(STDIN_FILENO) || argc != 3) {
    	syslog(LOG_NOTICE,
    	       "inappropriate use of PAM helper binary [UID=%d]",
               getuid());
    	sleep(10);	/* This should discourage/annoy the user. */
    	fprintf(stderr,
    	        "This binary is not designed for running in this way\n"
    	        "-- the system administrator has been informed\n");
    	return PAM_SYSTEM_ERR;
    }

    pw = getpwuid(getuid());
    if (pw == NULL) {
    	syslog(LOG_DEBUG, "no pw entry for current user, aborting.");
        return PAM_AUTH_ERR;
    }
    strncpy(user, pw->pw_name, sizeof(user) - 1);
    user[sizeof(user) - 1] = '\0';

    if (strcmp(user, argv[1])) {
        syslog(LOG_DEBUG, "current user does not match requested user, "
               "aborting.");
        return PAM_AUTH_ERR;
    }

    option = argv[2];
    if (strncmp(option, "verify", 8) == 0) {
        /* Unsupported. */
    	syslog(LOG_DEBUG, "unsupported option 'verify', aborting.");
        return PAM_AUTH_ERR;
    }
    /* We ignore any other options.  */

    /*
     * Read the password from stdin.
     */

    npass = read(STDIN_FILENO, pass, MAXPASS);

    /* Is it a valid password? */
    if (npass <= 0 || pass == NULL || pass[0] == '\0') {	
    	syslog(LOG_DEBUG, "no password supplied");
    } else if (npass >= MAXPASS) {
    	syslog(LOG_DEBUG, "password too long");
    } else {
    	/* Does pass agree with the official one? */
    	pass[npass] = '\0';	/* NUL terminate */
    	retval = auth_pam(user, pass);
    }
    memset(pass, '\0', MAXPASS);        /* clear memory of the password */

    /* return pass or fail */

    if (retval != PAM_SUCCESS) {
        return PAM_AUTH_ERR;
    } else {
        return PAM_SUCCESS;
    }
}

/*
 * vim: ts=4 sw=4 sts=4 et ai tw=80
 */
